from tvu._tvu import TVU
from tvu._compat import text


def bounded_int(minimum=None, maximum=None):

    class BoundedInt(TVU):

        TYPES = (int, float)

        def unify(self, value):
            if isinstance(value, float) and round(value) != value:
                self.error(u'a round number')
            return int(value)

        def validate(self, value):
            if minimum is not None and maximum is not None\
               and not (minimum <= value <= maximum):
                err_msg = \
                    u'an integer between %d and %d (inclusive)' % (minimum,
                                                                   maximum)
                self.error(err_msg)
            elif minimum is not None and value < minimum:
                err_msg = u'an integer greater or equal to %d' % (minimum,)
                self.error(err_msg)
            elif maximum is not None and value > maximum:
                err_msg = u'an integer lesser or equal to %d' % (maximum,)
                self.error(err_msg)

    return BoundedInt

PositiveInt = bounded_int(minimum=1)


class Text(TVU):

    TYPES = (text, bytes)

    def error(self, msg, soft=False):
        word = u'could' if soft else u'must'
        # skip adding unicode(value) to err_msg because there are problems
        # converting it to unicode
        raise ValueError((u'%s %s be ' + msg) % (self._variable_name, word))

    def unify(self, value):
        if isinstance(value, bytes):
            try:
                return value.decode('ascii')
            except UnicodeDecodeError:
                self.error(u'unicode text, or ascii-only bytestring')
        return value


class NonEmptyText(Text):

    def validate(self, value):
        if value is u'':
            self.error(u'non-empty string')


def iterable(elem_tvu=None):

    class Iterable(TVU):
        def type_check(self):
            value = self._value
            try:
                iter(value)
                return
            except TypeError:
                err_msg = \
                    u'%s must be iterable, not %s' % (self._variable_name,
                                                      text(self._value))
                raise TypeError(err_msg)

        def unify(self, value):
            if elem_tvu is None:
                return value

            result = []
            for i, elem in enumerate(value):
                new_var_name = u'%s[%d]' % (self._variable_name, i)
                validated_elem = elem_tvu(new_var_name).unify_validate(elem)
                result.append(validated_elem)
            return result

    return Iterable



import re
import sys
import unittest


class AssertRaisesContext(object):

    def __init__(self, test_case, exc_class, exc_msg, regex):
        self._test_case = test_case
        self._exc_class = exc_class
        self._exc_msg = exc_msg
        self._regex = regex

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, tb):
        if exc_value is None:
            test_err_msg = self._exc_class.__name__ + u' not raised'
            raise self._test_case.failureException(test_err_msg)
        if not isinstance(exc_value, self._exc_class):
            return False
        if self._exc_msg is not None and self._regex:
            regexp = re.compile(self._exc_msg)
            if not regexp.match(str(exc_value)):
                raise self._test_case.failureException(
                    '"%s" does not match "%s"' % (regexp.pattern,
                                                  str(exc_value)))
        elif self._exc_msg is not None:
            self._test_case.assertEqual(str(exc_value), self._exc_msg)
        return True


class TestCase(unittest.TestCase):

    def assertRaises(self, exc_class, exc_msg=None,
                     callableObj=None, regex=False, *args, **kwargs):
        if exc_msg is not None and sys.version_info < (3,):
            exc_msg = exc_msg.encode('ascii', 'replace').decode('ascii')
        if callableObj is None:
            return AssertRaisesContext(self, exc_class, exc_msg, regex)
        try:
            callableObj(*args, **kwargs)
        except exc_class as e:
            if exc_msg is not None:
                self.assertEqual(str(e), exc_msg)
        else:
            test_err_msg = exc_class.__name__ + u' not raised'
            raise self.failureException(test_err_msg)



# coding: utf-8
from test_utils import TestCase
from test_compat import text

import tvu
import tvu.tvus as tvus


class BoundedIntTest(TestCase):

    def unbounded_int_test(self):
        @tvu(x=tvus.bounded_int())
        def foo(x):
            return x

        self.assertEqual(foo(100), 100)
        self.assertEqual(foo(-100), -100)
        self.assertEqual(foo(1.), 1)
        self.assertTrue(isinstance(foo(1.), int))
        self.assertEqual(foo(-1.), -1)
        self.assertTrue(isinstance(foo(-1.), int))
        with self.assertRaises(TypeError, 'x must be int or float, not None'):
            foo(None)

        with self.assertRaises(ValueError,
                               'x must be a round number, not: 0.5'):
            foo(.5)

    def lower_bound_int_test(self):
        @tvu(x=tvus.bounded_int(minimum=0))
        def foo(x):
            return x

        self.assertEqual(foo(100), 100)
        self.assertEqual(foo(0), 0)
        self.assertEqual(foo(1.), 1)
        self.assertTrue(isinstance(foo(1.), int))
        with self.assertRaises(TypeError, 'x must be int or float, not None'):
            foo(None)

        with self.assertRaises(ValueError,
                               'x must be a round number, not: 0.5'):
            foo(.5)

        err_msg = 'x must be an integer greater or equal to 0, not: -1'
        with self.assertRaises(ValueError, err_msg):
            foo(-1)
        err_msg = 'x must be an integer greater or equal to 0, not: -1.0'
        with self.assertRaises(ValueError, err_msg):
            foo(-1.)

    def upper_bound_int_test(self):
        @tvu(x=tvus.bounded_int(maximum=0))
        def foo(x):
            return x

        self.assertEqual(foo(-100), -100)
        self.assertEqual(foo(0), 0)
        self.assertEqual(foo(-1.), -1)
        self.assertTrue(isinstance(foo(-1.), int))
        with self.assertRaises(TypeError, 'x must be int or float, not None'):
            foo(None)

        with self.assertRaises(ValueError,
                               'x must be a round number, not: 0.5'):
            foo(.5)

        err_msg = 'x must be an integer lesser or equal to 0, not: 1'
        with self.assertRaises(ValueError, err_msg):
            foo(1)
        err_msg = 'x must be an integer lesser or equal to 0, not: 1.0'
        with self.assertRaises(ValueError, err_msg):
            foo(1.)

    def bound_int_test(self):
        @tvu(x=tvus.bounded_int(minimum=0, maximum=10))
        def foo(x):
            return x

        self.assertEqual(foo(0), 0)
        self.assertEqual(foo(10), 10)
        self.assertEqual(foo(1.), 1)
        self.assertTrue(isinstance(foo(1.), int))
        with self.assertRaises(TypeError, 'x must be int or float, not None'):
            foo(None)

        with self.assertRaises(ValueError,
                               'x must be a round number, not: 0.5'):
            foo(.5)

        err_msg = 'x must be an integer between 0 and 10 (inclusive), not: 11'
        with self.assertRaises(ValueError, err_msg):
            foo(11)
        err_msg = \
            'x must be an integer between 0 and 10 (inclusive), not: -1.0'
        with self.assertRaises(ValueError, err_msg):
            foo(-1.)

    def positive_int_test(self):
        @tvu(x=tvus.PositiveInt)
        def foo(x):
            return x

        self.assertEqual(foo(100), 100)
        self.assertEqual(foo(1.), 1)
        self.assertTrue(isinstance(foo(1.), int))
        with self.assertRaises(TypeError, 'x must be int or float, not None'):
            foo(None)

        with self.assertRaises(ValueError,
                               'x must be a round number, not: 0.5'):
            foo(.5)

        err_msg = 'x must be an integer greater or equal to 1, not: -1'
        with self.assertRaises(ValueError, err_msg):
            foo(-1)
        err_msg = 'x must be an integer greater or equal to 1, not: 0.0'
        with self.assertRaises(ValueError, err_msg):
            foo(0.)


class IterableTest(TestCase):

    def iterable_test(self):
        @tvu(x=tvus.iterable())
        def foo(x):
            pass

        foo([])
        foo(iter({1, 2, 3}))
        with self.assertRaises(TypeError, 'x must be iterable, not None'):
            foo(None)

    def not_stealing_test(self):
        @tvu(x=tvus.iterable())
        def foo(x):
            return next(iter(x))

        self.assertEqual(foo([1, 2]), 1)

        def bar():
            yield 1
            yield 2

        self.assertEqual(foo(bar()), 1)

    def iterable_with_elem_test(self):
        @tvu(x=tvus.iterable(tvus.PositiveInt))
        def foo(x):
            pass

        foo([])
        foo(iter({1, 2, 3}))
        with self.assertRaises(TypeError, 'x must be iterable, not None'):
            foo(None)

        err_msg = "x[0] must be int or float, not 'bar'"
        with self.assertRaises(TypeError, err_msg):
            foo(['bar'])

        err_msg = \
            "x[0] must be an integer greater or equal to 1, not: 0"
        with self.assertRaises(ValueError, err_msg):
            foo([0])

        err_msg = "x[1] must be int or float, not 'bar'"
        with self.assertRaises(TypeError, err_msg):
            foo([1, 'bar'])

        err_msg = \
            "x[1] must be an integer greater or equal to 1, not: 0"
        with self.assertRaises(ValueError, err_msg):
            foo([1, 0])


class TextTest(TestCase):

    def text_test(self):
        @tvu(x=tvus.Text)
        def foo(x):
            return x

        self.assertEqual(foo(b''), u'')
        self.assertTrue(isinstance(foo(b''), text))
        self.assertEqual(foo(u''), u'')
        self.assertTrue(isinstance(foo(u''), text))
        self.assertEqual(foo(b'foo'), u'foo')
        self.assertTrue(isinstance(foo(b'foo'), text))
        self.assertEqual(foo(u'bar'), u'bar')
        self.assertTrue(isinstance(foo(u'bar'), text))

        err_msg = 'x must be %s or %s, not None' % (text.__name__,
                                                    bytes.__name__)
        with self.assertRaises(TypeError, err_msg):
            foo(None)

        err_msg = 'x must be unicode text, or ascii-only bytestring'
        with self.assertRaises(ValueError, err_msg):
            foo((u'żółw').encode('utf-8'))


class NonEmptyTextTest(TestCase):

    def non_empty_text_test(self):
        @tvu(x=tvus.NonEmptyText)
        def foo(x):
            return x

        self.assertEqual(foo(b'foo'), u'foo')
        self.assertTrue(isinstance(foo(b'foo'), type(u'')))  # compat
        self.assertEqual(foo(u'bar'), u'bar')
        self.assertTrue(isinstance(foo(u'bar'), type(u'')))  # compat

        err_msg = 'x must be %s or %s, not None' % (text.__name__,
                                                    bytes.__name__)
        with self.assertRaises(TypeError, err_msg):
            foo(None)

        err_msg = 'x must be unicode text, or ascii-only bytestring'
        with self.assertRaises(ValueError, err_msg):
            foo((u'żółw').encode('utf-8'))

        with self.assertRaises(ValueError, 'x must be non-empty string'):
            foo(b'')

        with self.assertRaises(ValueError, 'x must be non-empty string'):
            foo(u'')



from enum import Enum

import tvu
from tvu._compat import text, basestr
from test_utils import TestCase


class SampleEnum(Enum):
    one = 1
    two = 2
    three = 3


class TVUTest(TestCase):

    def basic_test(self):
        class IntTVU(tvu.TVU):
            TYPES = int,

        @tvu(x=IntTVU)
        def foo(x):
            return x

        self.assertEqual(foo(1), 1)
        with self.assertRaises(TypeError, 'x must be int, not None'):
            foo(None)

    def two_types_test(self):
        class IntStrTVU(tvu.TVU):
            TYPES = int, str

        @tvu(x=IntStrTVU)
        def foo(x):
            pass

        foo(1)
        foo('a')
        with self.assertRaises(TypeError, 'x must be int or str, not None'):
            foo(None)

    def multiple_types_test(self):
        class IntStrListTVU(tvu.TVU):
            TYPES = int, str, list

        @tvu(x=IntStrListTVU)
        def foo(x):
            pass

        foo(1)
        foo('a')
        foo([])
        with self.assertRaises(TypeError,
                               'x must be int, str or list, not None'):
            foo(None)

    def missing_validator_test(self):
        class IntTVU(tvu.TVU):
            TYPES = int,

        @tvu(x=IntTVU)
        def foo(x, y):
            pass

        foo(1, 2)
        foo(2, y=3)
        with self.assertRaises(TypeError,
                               'x must be int, not None'):
            foo(None, None)

    def multiple_validators_test(self):
        class IntTVU(tvu.TVU):
            TYPES = int,

        class StrTVU(tvu.TVU):
            TYPES = str,

        class FloatTVU(tvu.TVU):
            TYPES = float,

        @tvu(x=IntTVU, y=StrTVU, z=FloatTVU)
        def foo(x, y, z):
            pass

        foo(1, 'a', .0)

        with self.assertRaises(TypeError, "x must be int, not '1'"):
            foo('1', 'a', .0)
        with self.assertRaises(TypeError, 'y must be str, not 1'):
            foo(x=1, y=1, z=.0)
        with self.assertRaises(TypeError, "z must be float, not '1.0'"):
            foo(1, 'a', z='1.0')

    def unification_test(self):
        class NumberTVU(tvu.TVU):
            TYPES = float, int

            def unify(self, value):
                if isinstance(value, int):
                    return float(value)
                else:
                    return value

        @tvu(x=NumberTVU)
        def foo(x):
            return x

        self.assertEqual(foo(1), 1.)
        self.assertTrue(isinstance(foo(1), float))
        self.assertEqual(foo(1.), 1.)
        self.assertTrue(isinstance(foo(1.), float))
        with self.assertRaises(TypeError,
                               'x must be float or int, not None'):
            foo(None)

    def validation_test(self):
        class PositiveIntTVU(tvu.TVU):
            TYPES = int,

            def validate(self, value):
                if value <= 0:
                    self.error(u'positive number')

        @tvu(x=PositiveIntTVU)
        def foo(x):
            pass

        foo(1)
        with self.assertRaises(TypeError, 'x must be int, not None'):
            foo(None)

        with self.assertRaises(ValueError,
                               'x must be positive number, not: 0'):
            foo(0)

        with self.assertRaises(ValueError,
                               'x must be positive number, not: -1'):
            foo(-1)

    def no_types_test(self):
        class NullTVU(tvu.TVU):
            pass

        @tvu(x=NullTVU)
        def foo(x):
            pass

        foo(0)

    def enum_test(self):
        class SampleEnumTVU(tvu.EnumTVU):
            TYPES = SampleEnum,

        @tvu(x=SampleEnumTVU)
        def foo(x):
            return x.value

        self.assertEqual(foo(SampleEnum.one), 1)
        self.assertEqual(foo(SampleEnum.two), 2)
        self.assertEqual(foo(SampleEnum.three), 3)
        self.assertEqual(foo('one'), 1)
        self.assertEqual(foo('two'), 2)
        self.assertEqual(foo(u'three'), 3)

        with self.assertRaises(TypeError,
                               'x must be SampleEnum, not None'):
            foo(None)

        err_msg = "x could be SampleEnum's variant name, not: 'None'"
        with self.assertRaises(ValueError, err_msg):
            foo('None')

    def unicode_enum_test(self):
        class UnicodeEnum(text, Enum):
            foo = u'bar'

        @tvu(x=tvu.instance(UnicodeEnum, enum=True))
        def foo(x):
            return x.value

        self.assertEqual(foo(UnicodeEnum.foo), u'bar')
        self.assertEqual(foo('foo'), u'bar')

        err_msg = "x could be UnicodeEnum's variant name, not: " + repr(u'bar')
        with self.assertRaises(ValueError, err_msg):
            foo(u'bar')

        err_msg = "x could be UnicodeEnum's variant name, not: 'bar'"
        with self.assertRaises(ValueError, err_msg):
            foo('bar')

    def str_enum_test(self):
        class StrEnum(str, Enum):
            foo = 'bar'

        @tvu(x=tvu.instance(StrEnum, enum=True))
        def foo(x):
            return x.value

        self.assertEqual(foo(StrEnum.foo), 'bar')
        self.assertEqual(foo('foo'), 'bar')

        err_msg = "x could be StrEnum's variant name, not: 'bar'"
        with self.assertRaises(ValueError, err_msg):
            foo('bar')

    def instance_test(self):
        @tvu(x=tvu.instance(int))
        def foo(x):
            return x

        self.assertEqual(foo(1), 1)
        with self.assertRaises(TypeError, 'x must be int, not None'):
            foo(None)

    def instance_enum_test(self):
        @tvu(x=tvu.instance(SampleEnum, enum=True))
        def foo(x):
            return x.value

        self.assertEqual(foo(SampleEnum.one), 1)
        self.assertEqual(foo(SampleEnum.two), 2)
        self.assertEqual(foo(SampleEnum.three), 3)
        self.assertEqual(foo('one'), 1)
        self.assertEqual(foo('two'), 2)
        self.assertEqual(foo('three'), 3)

        with self.assertRaises(TypeError,
                               'x must be SampleEnum, not None'):
            foo(None)

        err_msg = "x could be SampleEnum's variant name, not: 'None'"
        with self.assertRaises(ValueError, err_msg):
            foo('None')

    def nullable_basic_test(self):
        @tvu(x=tvu.nullable(tvu.instance(int)))
        def foo(x):
            return x

        self.assertEqual(foo(1), 1)
        self.assertEqual(foo(None), None)
        with self.assertRaises(TypeError, "x must be int or None, not 'a'"):
            foo('a')

    def nullable_complex_test(self):
        class PositiveNumberTVU(tvu.TVU):
            TYPES = (int, float)

            def unify(self, value):
                if isinstance(value, int):
                    return float(value)
                else:
                    return value

            def validate(self, value):
                if value <= 0.:
                    self.error(u'positive number')

        @tvu(x=tvu.nullable(PositiveNumberTVU))
        def foo(x):
            return x if x else 0.

        self.assertEqual(foo(1), 1.)
        self.assertEqual(foo(2), 2.)
        self.assertEqual(foo(None), 0.)

        with self.assertRaises(TypeError,
                               "x must be int, float or None, not 'a'"):
            foo('a')

        with self.assertRaises(ValueError,
                               'x must be None or positive number, not: 0'):
            foo(0)

        with self.assertRaises(ValueError,
                               'x must be None or positive number, not: 0.0'):
            foo(0.)

    def nullable_soft_error_test(self):
        @tvu(x=tvu.nullable(tvu.instance(SampleEnum, enum=True)))
        def foo(x):
            pass

        err_msg = "x could be None or SampleEnum's variant name, not: 'bar'"
        with self.assertRaises(ValueError, err_msg):
            foo('bar')



# coding: utf-8
import inspect

import enum

from tvu._compat import basestr


class TVU(object):

    TYPES = (object,)

    def __init__(self, variable_name=None):
        self._variable_name = \
            variable_name or self.__class__.__name__ + u'() argument'

    def _type_name(self, type_):
        return type_.__name__

    def type_check(self):
        value = self._value
        types = self.__class__.TYPES
        if isinstance(value, types):
            return

        if len(types) == 1:
            possible_types_string = self._type_name(types[0])
        else:
            possible_types_string = \
                u', '.join([self._type_name(type_) for type_ in types[:-1]]) \
                + u' or ' + self._type_name(types[-1])

        err_msg = u'%s must be %s, not %s' % (self._variable_name,
                                              possible_types_string,
                                              repr(self._value))
        raise TypeError(err_msg)

    def unify_validate(self, value):
        self._value = value
        self.type_check()
        unified_value = self.unify(self._value)
        self.validate(unified_value)
        return unified_value

    def error(self, msg, soft=False):
        word = u'could' if soft else u'must'
        full_msg = (u'%s %s be ' + msg + u', not: %s') % \
            (self._variable_name, word, repr(self._value))
        raise ValueError(full_msg)

    def unify(self, value):
        return value

    def validate(self, value):
        pass


def validate_and_unify(**arg_validators):
    def wrapper(fun):
        spec_args, spec_varargs, spec_keywords, spec_defaults = \
            inspect.getargspec(fun)

        def inner_wrapper(*args, **kwargs):
            args_values = inspect.getcallargs(fun, *args, **kwargs)
            for arg in args_values:
                try:
                    validator = arg_validators[arg]
                except KeyError:
                    pass
                else:
                    args_values[arg] = \
                        validator(arg).unify_validate(args_values[arg])

            new_args = []

            # TODO: support default arguments
            try:
                for arg in spec_args:
                    new_args.append(args_values.pop(arg))
                if spec_varargs is not None:
                    new_args += args_values.pop(spec_varargs)
            except KeyError:
                raise TypeError()
            if spec_keywords is not None:
                kwargs = args_values.pop(spec_keywords)
                for key, val in kwargs.items():
                    args_values[key] = val

            return fun(*new_args, **args_values)
        return inner_wrapper
    return wrapper


class EnumTVU(TVU):

    def _get_enum_type(self):
        for type_ in self.TYPES:
            if issubclass(type_, enum.Enum):
                return type_

    def type_check(self):
        value = self._value
        if isinstance(value, basestr) and not isinstance(value, enum.Enum):
            enum_type = self._get_enum_type()
            try:
                self._value = getattr(enum_type, value)
            except AttributeError:
                msg = enum_type.__name__ + u"'s variant name"
                self.error(msg, soft=True)
            return
        super(EnumTVU, self).type_check()


def instance(class_, enum=False):
    base = EnumTVU if enum else TVU

    class InstanceTVU(base):
        TYPES = (class_,)

    return InstanceTVU


def nullable(tvu):

    class NullableTVU(tvu):
        TYPES = tvu.TYPES + (type(None),)

        def _type_name(self, type_):
            if isinstance(None, type_):
                return 'None'
            else:
                return super(NullableTVU, self)._type_name(type_)

        def unify(self, value):
            if value is None:
                return value
            return super(NullableTVU, self).unify(value)

        def validate(self, value):
            if value is None:
                return
            super(NullableTVU, self).validate(value)

        def error(self, err_msg, soft=False):
            err_msg = u'None or ' + err_msg
            return super(NullableTVU, self).error(err_msg, soft)

    return NullableTVU



# -*- coding: utf-8 -*-
import sys
try:
    from setuptools import setup
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup


__version__ = '0.2.2'


tests_require = ['nose == 1.3.7']
install_require = []
if sys.version_info < (3, 4):
    install_require.append('enum34 == 1.1.6')

short_descr = \
    'Library for typechecking/validation/unification of function arguments.'
download_url = 'http://github.com/Paczesiowa/tvu/tarball/v' + __version__

setup(name='tvu',
      version=__version__,
      packages=['tvu'],
      description=short_descr,
      long_description=open('README.md', 'r').read(),
      author=u'Bartek Ćwikłowski',
      author_email='paczesiowa@gmail.com',
      url='https://github.com/Paczesiowa/tvu',
      download_url=download_url,
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'License :: Public Domain',
                   'Operating System :: POSIX',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.6'],
      license='UNLICENSE/Public Domain',
      keywords=['typecheck', 'decorator'],
      zip_safe=True,
      install_requires=install_require,
      test_suite='nose.collector',
      tests_require=tests_require)



import pkg_resources
import sys


class TVUModule(object):

    def __init__(self, name):
        self.__name__ = name

        import tvu._tvu
        import tvu.tvus

        self.tvus = tvu.tvus

        self.TVU = tvu._tvu.TVU
        self.EnumTVU = tvu._tvu.EnumTVU

        self.instance = tvu._tvu.instance
        self.nullable = tvu._tvu.nullable

        self._validate_and_unify = tvu._tvu.validate_and_unify

        try:
            self.__version__ = pkg_resources.require('tvu')[0].version
        except pkg_resources.DistributionNotFound:
            self.__version__ = 'dev'

        self.__file__ = __file__

    def __call__(self, **kwargs):
        return self._validate_and_unify(**kwargs)


sys.modules[__name__] = TVUModule(__name__)
